package main

import (
	"database/sql"
	"fmt"
	_ "github.com/SAP/go-hdb/driver"
	_ "github.com/lib/pq"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	"log"
	"net/http"
	"os"
	"sync"
	"time"
)

var (
	driverName = "hdb"
	hdbDsnSysFmt     = "hdb://SYSTEM:%s@%s:%s"
	//big int is not supported in golang, use int instead
	sqlMem = "select TO_INTEGER(INSTANCE_TOTAL_MEMORY_USED_SIZE/1024/1024/1024) as \"usedMem\", TO_INTEGER(ALLOCATION_LIMIT/1024/1024/1024) as \"alloLimitMem\", " +
		"(TO_INTEGER(ALLOCATION_LIMIT/1024/1024/1024) - TO_INTEGER(INSTANCE_TOTAL_MEMORY_USED_SIZE/1024/1024/1024)) as \"allocateMem\" from M_HOST_RESOURCE_UTILIZATION"
	pgHost     = os.Getenv("PG_HOST")
	pgPort     = os.Getenv("PG_PORT")
	pgUser     = os.Getenv("PG_USER")
	pgPassword = os.Getenv("PG_PASSWORD")
	pgDbname   = os.Getenv("PG_DB")
	hanaName   = os.Getenv("HANA_NAME")
	hanaIp   = os.Getenv("HANA_IP")
	hanaPort   = os.Getenv("HANA_PORT")
	hanaSysPassword   = os.Getenv("HANA_SYSPASSWD")
	noDataZero   = os.Getenv("NO_DATA_ZERO")
)

var (
	dbinstances = []HanaInstance{}
	usedMemDesc = prometheus.NewDesc(
		"usedMem",
		"hana used memory",
		[]string{"name"}, nil,
	)
	allocateMemDesc = prometheus.NewDesc(
		"allocateMem",
		"hana allocatable memory",
		[]string{"name"}, nil,
	)
)

func fetchData(hanaInst HanaInstance, results chan<- prometheus.Metric,wg *sync.WaitGroup) {
	defer wg.Done()

	fmt.Println(hanaInst)
	//from hana
	name := hanaInst.name
	ip := hanaInst.ip
	port := hanaInst.port

	hdbUrl := fmt.Sprintf(hdbDsnSysFmt,hanaSysPassword,ip,port)
	db, err := sql.Open(driverName, hdbUrl)
	if (err != nil) {
		fmt.Println(err)
		return
	}
	defer db.Close()

	var usedMem,alloLimitMem,allocateMem float64
	//allocateMem = 0.0

	fmt.Println(sqlMem)
	rows, err := db.Query(sqlMem)
	if (err == nil) {
		defer rows.Close()
		for rows.Next() {
			err = rows.Scan(&usedMem,&alloLimitMem,&allocateMem)
			CheckError(err)
			fmt.Println("get hana data:",usedMem,alloLimitMem,allocateMem)
		}
		//return
	} else if (noDataZero == "true"){
		fmt.Println("noDataZero,hana query error:",name,err)
	} else {
		fmt.Println("hana query error:",name,err)
		return
	}

	results <- prometheus.MustNewConstMetric(usedMemDesc, prometheus.GaugeValue,usedMem,name)
	results <- prometheus.MustNewConstMetric(allocateMemDesc, prometheus.GaugeValue,allocateMem,name)
}

func CheckError(err error) {
	if err != nil {
		log.Fatal(err)//throw exception
	}
}

func PanicError(err error) {
	if err != nil {
		panic(err)
	}
}

type HanaInstance struct {
	name string
	ip string
	port string
}

type ExCollector struct {
}

func (arc ExCollector) Describe(ch chan<- *prometheus.Desc) {
	ch <- usedMemDesc
	ch <- allocateMemDesc
}

func (arc ExCollector) Collect(ch chan<- prometheus.Metric)  {
	fmt.Println("Scraping cmdb HANA metrics data...")
	//get from CMDB

	var wg sync.WaitGroup
	basicChan := make(chan prometheus.Metric, 10000)

	for _,hanaInst := range dbinstances {
		wg.Add(1)
		go fetchData(hanaInst,basicChan, &wg)
	}
	wg.Wait()
	close(basicChan)

	for result := range basicChan {
		ch <- result
	}
	now := time.Now()
	fmt.Println(now)
}

func getHanaInstances(pgDb *sql.DB) error {
	sqlHana := "select \"name\",frpc_ip,frpc_hana_port from hana_info"
	rows, err := pgDb.Query(sqlHana)
	CheckError(err)
	defer rows.Close()

	fmt.Println(rows)
	for rows.Next() {
		var hana_name, ip, port string
		err = rows.Scan(&hana_name, &ip, &port)
		CheckError(err)
		dbinstances = append(dbinstances, HanaInstance{hana_name,ip,port})
	}
	fmt.Println(dbinstances)

	return nil
}

func main() {
	fmt.Println("hana exporter begin")
	now := time.Now()
	fmt.Println(now)

	if hanaName == "" || hanaIp == "" || hanaPort == "" {
		//pg
		fmt.Println("fetch data from CMDB")
		psqlconn := fmt.Sprintf("host=%s port=%s user=%s password=%s dbname=%s sslmode=disable", pgHost, pgPort, pgUser, pgPassword, pgDbname)
		fmt.Println(psqlconn)
		// open database
		pgDb, err := sql.Open("postgres", psqlconn)
		CheckError(err)
		// close database
		defer pgDb.Close()
		// check db
		err = pgDb.Ping()
		CheckError(err)
		fmt.Println("Pg Connected!")

		err1 := getHanaInstances(pgDb)
		CheckError(err1)
	} else {
		fmt.Println("fetch data from specified hana")
		dbinstances = append(dbinstances, HanaInstance{hanaName,hanaIp,hanaPort})
	}
	arc := ExCollector{}
	prometheus.MustRegister(arc)
	http.Handle("/metrics", promhttp.Handler())
	log.Fatal(http.ListenAndServe(":9000", nil))
	fmt.Println("done")
}
